use chrono::{DateTime, Utc};
use serde::{Serialize, Serializer};

#[derive(Debug, Clone, Serialize)]
pub struct User {
    pub id: i32,
    pub username: String,
    #[serde(skip_serializing)]
    pub password: String,
    pub role: Role,

    pub created_at: DateTime<Utc>,
    pub last_active: DateTime<Utc>,
}

#[derive(sqlx::Type, Debug, Copy, Clone, PartialEq, Eq)]
#[repr(i32)]
pub enum Role {
    Viewer = 1,
    Editor = 2,
    Manager = 3,
    Admin = 4,
}

impl Serialize for Role {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        serializer.serialize_str(match self {
            Role::Viewer => "viewer",
            Role::Editor => "editor",
            Role::Manager => "manager",
            Role::Admin => "admin",
        })
    }
}

impl From<i32> for Role {
    fn from(value: i32) -> Self {
        // From<i32> is only used for database mapping, so panicking on invalid values is okay
        match value {
            1 => Self::Viewer,
            2 => Self::Editor,
            3 => Self::Manager,
            4 => Self::Admin,
            _ => unreachable!("Role should never be outside 1-4"),
        }
    }
}

impl Into<i32> for Role {
    fn into(self) -> i32 {
        match self {
            Self::Viewer => 1,
            Self::Editor => 2,
            Self::Manager => 3,
            Self::Admin => 4,
        }
    }
}

impl TryFrom<&str> for Role {
    type Error = eyre::Error;

    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value {
            "viewer" => Ok(Self::Viewer),
            "editor" => Ok(Self::Editor),
            "manager" => Ok(Self::Manager),
            "admin" => Ok(Self::Admin),
            _ => Err(eyre::eyre!("Invalid role string value")),
        }
    }
}

impl Into<&str> for Role {
    fn into(self) -> &'static str {
        match self {
            Self::Viewer => "viewer",
            Self::Editor => "editor",
            Self::Manager => "manager",
            Self::Admin => "admin",
        }
    }
}

impl Into<String> for Role {
    fn into(self) -> String {
        match self {
            Self::Viewer => "viewer".into(),
            Self::Editor => "editor".into(),
            Self::Manager => "manager".into(),
            Self::Admin => "admin".into(),
        }
    }
}
