use std::{collections::BTreeMap, sync::Arc};

use axum::{
    body::{BoxBody, Full},
    extract::{Path, State},
    http::header::{CONTENT_TYPE, SET_COOKIE},
    response::{IntoResponse, Response},
    Form,
};
use serde::Deserialize;
use sqlx::{query, query_as};
use tracing::{error, info};

use crate::{
    extractor::{ExtractUserToken, TOKEN_COOKIE_NAME},
    model::user::{Role, User},
    state::AppState,
    templates::{Page, PageData},
    token::Claims,
};

pub async fn get_user(
    State(state): State<Arc<AppState>>,
    Path(user_id): Path<i32>,
) -> impl IntoResponse {
    let user = match query_as!(User, r#"SELECT * FROM users WHERE id = $1"#, user_id)
        .fetch_one(&state.pool)
        .await
    {
        Ok(user) => user,
        Err(err) => {
            error!("Getting user: {}", err);
            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Internal server error".into()),
                    ..Default::default()
                },
            ));
        }
    };

    Page::new(state.hbs.render(
        "error.hbs",
        &PageData {
            user: Some(user),
            ..Default::default()
        },
    ))
}

pub async fn get_user_new(
    State(state): State<Arc<AppState>>,
    ExtractUserToken(user): ExtractUserToken,
) -> impl IntoResponse {
    // If there's already an authenticated user, check if they're an admin
    // Admins can create new users, others can't
    if let Some(user) = user {
        tracing::debug!("User is authenticated");

        if user.role == Role::Admin {
            tracing::debug!("Authenticated user is admin");

            return Page::new(state.hbs.render(
                "new_user.hbs",
                &PageData {
                    authed_user: Some(user),
                    ..Default::default()
                },
            ));
        }

        tracing::debug!("Authenticated user is not an admin");
        return Page::new(state.hbs.render::<PageData>(
            "error.hbs",
            &PageData {
                authed_user: Some(user),
                message: Some("You are not an admin".into()),
                ..Default::default()
            },
        ));
    }

    tracing::debug!("There is no authenticated user");

    // Else, let the user create a new account, but only if no accounts exist yet
    let count = match query!(r#"SELECT COUNT(id) FROM users"#)
        .fetch_one(&state.pool)
        .await
    {
        Ok(r) => r.count.unwrap_or(0),
        Err(err) => {
            error!("Getting user count: {}", err);
            return Page::new(
                state
                    .hbs
                    .render::<PageData>("error.hbs", &Default::default()),
            );
        }
    };

    tracing::debug!("Current user count: {}", count);

    let mut map = BTreeMap::new();
    map.insert("count", count);

    Page::new(state.hbs.render("new_user.hbs", &map))
}

#[derive(Deserialize)]
pub struct UserNew {
    pub username: String,
    pub password: String,
    pub password2: String,
}

pub async fn post_user_new(
    State(state): State<Arc<AppState>>,
    ExtractUserToken(existing_user): ExtractUserToken,
    Form(form_data): Form<UserNew>,
) -> Response<BoxBody> {
    if existing_user.is_none() {
        let count = match query!(r#"SELECT COUNT(id) FROM users"#)
            .fetch_one(&state.pool)
            .await
        {
            Ok(r) => r.count.unwrap_or(0),
            Err(err) => {
                error!("Getting user count: {}", err);

                return Page::new(state.hbs.render(
                    "error.hbs",
                    &PageData {
                        message: Some("Internal server error".into()),
                        ..Default::default()
                    },
                ))
                .into_response();
            }
        };

        if count != 0 {
            info!("");

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("You need to be logged in to create new users".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    } else if let Some(user) = existing_user.clone() {
        if user.role != Role::Admin {
            info!(
                user_id = user.id,
                "User tried to access POST /users/new despite not being admin"
            );

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Only admins can create new users".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    }

    if form_data.password != form_data.password2 {
        return Page::new(state.hbs.render(
            "new_user.hbs",
            &PageData {
                message: Some("Passwords don't match".into()),
                ..Default::default()
            },
        ))
        .into_response();
    }

    let hashed_password = match bcrypt::hash(form_data.password, 12) {
        Ok(hash) => hash,
        Err(err) => {
            error!("Hashing password: {}", err);

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Internal server error".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    };

    let role = match existing_user.is_some() {
        true => Role::Viewer,
        false => Role::Admin,
    };

    let user = match query_as!(
        User,
        r#"INSERT INTO users (username, password, role) VALUES ($1, $2, $3) RETURNING *"#,
        form_data.username,
        hashed_password,
        <Role as Into<i32>>::into(role),
    )
    .fetch_one(&state.pool)
    .await
    {
        Ok(u) => u,
        Err(err) => {
            error!("Creating user: {}", err);

            return Page::new(state.hbs.render(
                "error.hbs",
                &PageData {
                    message: Some("Internal server error".into()),
                    ..Default::default()
                },
            ))
            .into_response();
        }
    };

    if existing_user.is_some() {
        // Render template with new user info (ID, username)

        return Page::new(state.hbs.render(
            "new_user_confirm.hbs",
            &PageData {
                user: existing_user,
                ..Default::default()
            },
        ))
        .into_response();
    }

    // Else, create cookie
    let token = match Claims::new(&user).encode(&state.encoding_key) {
        Ok(token) => token,
        Err(err) => {
            error!("Encoding token: {}", err);

            return Page::new(state.hbs.render("error.hbs", &PageData{ 
                message: Some("Internal server error. Note: your account was created successfully, please log in manually.".into()),
                ..Default::default()
            })).into_response();
        }
    };

    let content = match state.hbs.render(
        "new_user_confirm.hbs",
        &PageData {
            authed_user: Some(user.clone()),
            user: Some(user),
            message: Some("Your account has been created!".into()),
        },
    ) {
        Ok(content) => content,
        Err(err) => {
            error!("Rendering content: {}", err);

            return Page::new(state.hbs.render("error.hbs", &PageData{ 
                message: Some("Internal server error. Note: your account was created successfully, please log in manually.".into()),
                ..Default::default()
            })).into_response();
        }
    };

    return match Response::builder()
        .header(SET_COOKIE, format!("{}={}", TOKEN_COOKIE_NAME, token))
        .header(CONTENT_TYPE, "text/html; charset=utf-8")
        .body(Full::from(content))
    {
        Ok(resp) => resp.into_response(),
        Err(err) => {
            error!("Building response: {}", err);

            return Page::new(state.hbs.render("error.hbs", &PageData{ 
                message: Some("Internal server error. Note: your account was created successfully, please log in manually.".into()),
                ..Default::default()
            })).into_response();
        }
    };
}
